==> delayed_reactant_labeling-0.1.7-py3-none-any.whl: delayed_reactant_labeling/optimize_rate_constants.py <==
from abc import ABC, abstractmethod
from typing import Optional, Callable

from scipy.optimize import minimize
from sklearn.metrics import mean_absolute_error
import pandas as pd
import numpy as np
from tqdm import tqdm
from datetime import datetime
import os


class JSON_log:
    def __init__(self, path, mode="new"):
        self._path = path
        exists = os.path.isfile(path)

        if mode == "new":
            # create a new file
            if exists:
                raise FileExistsError("File already exists. To replace it use mode='replace'")
            with open(self._path, "w") as _:
                pass

        elif mode == "append":
            # append to the file
            if not exists:
                raise ValueError("File does not exist. Use mode='new' to create it.")

        elif mode == "replace":
            # replace the file
            with open(self._path, "w") as _:
                pass

    def log(self, data: pd.Series):
        data["datetime"] = datetime.now().strftime('%d/%m/%Y %H:%M:%S')
        with open(self._path, "a") as f:
            f.write(data.to_json() + "\n")


class OptimizerProgress:
    def __init__(self, path: str):
        # read the meta data
        self.metadata = pd.read_json(f"{path}/settings_info.json", lines=True).iloc[0, :]
        self.x_description = np.array(self.metadata["x_description"])

        # read the optimization log
        df = pd.read_json(f"{path}/optimization_log.json", lines=True)
        self.n_dimensions = len(self.x_description)
        self.n_iterations = len(df)

        self._all_X: pd.DataFrame = pd.DataFrame(list(df.loc[:, "x"]), columns=self.x_description)
        self._all_errors: pd.Series = df["error"]
        self._all_ratios: pd.Series = df["ratio"]
        self._all_times: pd.Series = df["datetime"]

        simplex = np.full((self.n_dimensions+1, self.n_dimensions), np.nan)
        sorted_errors = self._all_errors.sort_values(ascending=True)
        for n, index in enumerate(sorted_errors[:self.n_dimensions+1].keys()):
            simplex[n, :] = self._all_X.iloc[index, :].to_numpy()  # underscored variant so that no copying is required

        best_iteration_index = sorted_errors.index[0]

        self._best_X: pd.Series = pd.Series(self.all_X.loc[best_iteration_index, :], index=self.x_description)
        self.best_error: float = self.all_errors[best_iteration_index]
        self.best_ratio: float = self.all_ratios[best_iteration_index]

    @property
    def best_X(self):
        return self._best_X.copy()

    @property
    def all_X(self):
        return self._all_X.copy()

    @property
    def all_errors(self):
        return self._all_errors.copy()

    @property
    def all_time_stamps(self):
        return self._all_times.copy()

    @property
    def all_ratios(self):
        return self._all_ratios.copy()


class RateConstantOptimizerTemplate(ABC):
    def __init__(self, raw_weights: dict[str, float], experimental: pd.DataFrame) -> None:
        """
        Initializes the Rate constant optimizer class.
        :param raw_weights: dictionary containing str patterns as keys and the weight as values.
        The str patterns will be searched for in the keys of the results from the 'calculate_curves' function.
        The given weight will lower corresponding errors.
        :param experimental: The experimental data
        """
        self.raw_weights = raw_weights
        self.weights = None

        # initialize all curves for the experimental (true) values.
        self.experimental_curves = self.calculate_curves(experimental)

        # check if any of the curves are potentially problematic
        for curve_description, curve in self.experimental_curves.items():
            if any(np.isnan(curve)):
                raise ValueError(f"Experimental data curve for {curve_description} contains NaN values.")

    @abstractmethod
    def create_prediction(self, x: np.ndarray, x_description: list[str]) -> tuple[pd.DataFrame, float]:
        """
        Create a prediction of the system, given a set of parameters.
        :param x: Contains all parameters, which are to be optimized.
        Definitely includes are rate constants.
        :param x_description: Contains a description of each parameter.
        :return: Predicted values of the concentration for each chemical, as a function of time.
                 Predicted ratio of compounds (enantiomeric ratio).
        """
        pass

    @staticmethod
    @abstractmethod
    def calculate_curves(data: pd.DataFrame) -> dict[str, pd.Series]:
        """
        Calculate the curves corresponding to the data (either experimental or predicted).
        The experimental curves will only be calculated once and are stored for subsequent use.
        Internally, the experimental and predicted curves will be compared against each other to determine the error.
        :param data: The data from which the curves should be calculated
        :return: dict containing a description of each curve, and the corresponding curve.
        """

    def calculate_error_functions(self, prediction: pd.DataFrame, metric: Optional[Callable] = None) -> pd.Series:
        """
        Calculate the error caused by each error function.
        :param prediction: The predicted concentrations.
        :param metric: The error metric which should be used, defaults to the mean absolute error.
        :return: The unweighted errors of each error function.
        """
        if metric is None:
            metric = mean_absolute_error

        curves_predicted = self.calculate_curves(prediction)
        error = {}
        for curve_description, curve_prediction in curves_predicted.items():
            error[curve_description] = metric(
                y_true=self.experimental_curves[curve_description],
                y_pred=curve_prediction)

        return pd.Series(error)

    def weigh_errors(self, errors: pd.Series, ) -> pd.Series:
        """
        weighs the errors
        :param errors: unweighted errors
        :return: weighed errors
        """
        assert isinstance(errors, pd.Series)
        if self.weights is None:
            weights = np.ones(errors.shape)
            for description, weight in self.raw_weights.items():
                index = errors.index.str.contains(description)
                if len(index) == 0:
                    raise ValueError(f"no matches were found for {description}")
                weights[index] = weights[index] * weight
            self.weights = weights

        return errors * self.weights

    def calculate_total_error(self, errors: pd.Series) -> float:
        """
        weighs and sums the errors.
        :param errors: unweighted errors
        :return: weighed total error
        """
        return self.weigh_errors(errors).sum(skipna=False)

    def optimize(self,
                 x0: np.ndarray,
                 x_description: list[str],
                 bounds: list[tuple[float, float]],
                 path: str,
                 metadata: Optional[dict],
                 maxiter: float = 50000,
                 resume_from_simplex=None
                 ) -> None:
        """
        Optimizes the system, utilizing a nelder-mead algorithm.
        :param x0: Parameters which are to be optimized. Always contain the rate constants.
        :param x_description: Description of each parameter.
        :param bounds: A list containing tuples, which in turn contain the lower and upper bound for each parameter.
        :param path: Where the solution should be stored.
        :param metadata: The metadata that should be saved alongside the solution.
        :param maxiter: The maximum number of iterations.
        :param resume_from_simplex: When a simplex is given, the solution starts here.
        It can be used to resume the optimization process.
        """
        # enable logging of all information retrieved from the system
        log_path = f"{path}/optimization_log.json"
        if resume_from_simplex is None:  # new optimization progres
            logger = JSON_log(log_path)
            metadata_extended = {
                "raw_weights": self.raw_weights,
                "x0": x0,
                "x_description": x_description,
                "bounds": bounds,
                "maxiter": maxiter
            }
            if metadata is None:
                pass
            else:
                # overwrites the default meta data values
                for key, value in metadata.items():
                    metadata_extended[key] = value
            meta_data_log = JSON_log(f"{path}/settings_info.json", mode="new")
            meta_data_log.log(pd.Series(metadata_extended))
        else:
            logger = JSON_log(log_path, mode="append")

        def optimization_step(x):
            """The function is given a set of parameters by the Nelder-Mead algorithm.
            Proceeds to calculate the corresponding prediction and its total error.
            The results are stored in a log before the error is returned to the optimizer."""

            prediction, predicted_compound_ratio = self.create_prediction(x, x_description)
            errors = self.calculate_error_functions(prediction)
            total_error = self.calculate_total_error(errors)

            logger.log(pd.Series([x, total_error, predicted_compound_ratio], index=["x", "error", "ratio"]))
            return total_error

        def update_tqdm(_):
            """update the progress bar"""
            pbar.update(1)

        with tqdm(total=maxiter, miniters=25) as pbar:
            # the minimization process is stored within the log, containing all x's and errors.
            minimize(fun=optimization_step,
                     x0=x0,
                     method="Nelder-Mead",
                     bounds=bounds,
                     callback=update_tqdm,
                     options={"maxiter": maxiter, "disp": True, "adaptive": True, "return_all": False,
                              "initial_simplex": resume_from_simplex})

    @staticmethod
    def load_optimization_progress(path: str) -> OptimizerProgress:
        """
        Loads in the data from the log files.
        :param path: Folder in which the optimization_log.json and settings_info.json files can be found.
        :return: OptimizerProgress instance which contains all information that was logged.
        """
        return OptimizerProgress(path)

==> delayed_reactant_labeling-0.1.8-py3-none-any.whl: delayed_reactant_labeling/optimize_rate_constants.py <==
from abc import ABC, abstractmethod
from typing import Optional, Callable

from scipy.optimize import minimize
from sklearn.metrics import mean_absolute_error
import pandas as pd
import numpy as np
from tqdm import tqdm
from datetime import datetime
import os


class JSON_log:
    def __init__(self, path, mode="new"):
        self._path = path
        exists = os.path.isfile(path)

        if mode == "new":
            # create a new file
            if exists:
                raise FileExistsError("File already exists. To replace it use mode='replace'")
            with open(self._path, "w") as _:
                pass

        elif mode == "append":
            # append to the file
            if not exists:
                raise ValueError("File does not exist. Use mode='new' to create it.")

        elif mode == "replace":
            # replace the file
            with open(self._path, "w") as _:
                pass

    def log(self, data: pd.Series):
        data["datetime"] = datetime.now().strftime('%d/%m/%Y %H:%M:%S')
        with open(self._path, "a") as f:
            f.write(data.to_json() + "\n")


class OptimizerProgress:
    def __init__(self, path: str):
        # read the meta data
        self.metadata = pd.read_json(f"{path}/settings_info.json", lines=True).iloc[0, :]
        self.x_description = np.array(self.metadata["x_description"])

        # read the optimization log
        df = pd.read_json(f"{path}/optimization_log.json", lines=True)
        self.n_dimensions = len(self.x_description)
        self.n_iterations = len(df)

        self._all_X: pd.DataFrame = pd.DataFrame(list(df.loc[:, "x"]), columns=self.x_description)
        self._all_errors: pd.Series = df["error"]
        self._all_ratios: pd.Series = df["ratio"]
        self._all_times: pd.Series = df["datetime"]

        simplex = np.full((self.n_dimensions+1, self.n_dimensions), np.nan)
        sorted_errors = self._all_errors.sort_values(ascending=True)
        for n, index in enumerate(sorted_errors[:self.n_dimensions+1].keys()):
            simplex[n, :] = self._all_X.iloc[index, :].to_numpy()  # underscored variant so that no copying is required

        best_iteration_index = sorted_errors.index[0]

        self._best_X: pd.Series = pd.Series(self.all_X.loc[best_iteration_index, :], index=self.x_description)
        self.best_error: float = self.all_errors[best_iteration_index]
        self.best_ratio: float = self.all_ratios[best_iteration_index]

    @property
    def best_X(self):
        return self._best_X.copy()

    @property
    def all_X(self):
        return self._all_X.copy()

    @property
    def all_errors(self):
        return self._all_errors.copy()

    @property
    def all_time_stamps(self):
        return self._all_times.copy()

    @property
    def all_ratios(self):
        return self._all_ratios.copy()


class RateConstantOptimizerTemplate(ABC):
    def __init__(self, raw_weights: dict[str, float], experimental: pd.DataFrame, metric: Optional[Callable[[np.ndarray, np.ndarray], float]]) -> None:
        """
        Initializes the Rate constant optimizer class.
        :param raw_weights: dictionary containing str patterns as keys and the weight as values.
        The str patterns will be searched for in the keys of the results from the 'calculate_curves' function.
        The given weight will lower corresponding errors.
        :param experimental: The experimental data
        """
        self.raw_weights = raw_weights
        self.weights = None

        # initialize all curves for the experimental (true) values.
        self.experimental_curves = self.calculate_curves(experimental)

        if metric is None:
            metric = mean_absolute_error
        self.metric = metric

        # check if any of the curves are potentially problematic
        for curve_description, curve in self.experimental_curves.items():
            if any(np.isnan(curve)):
                raise ValueError(f"Experimental data curve for {curve_description} contains NaN values.")

    @abstractmethod
    def create_prediction(self, x: np.ndarray, x_description: list[str]) -> tuple[pd.DataFrame, float]:
        """
        Create a prediction of the system, given a set of parameters.
        :param x: Contains all parameters, which are to be optimized.
        Definitely includes are rate constants.
        :param x_description: Contains a description of each parameter.
        :return: Predicted values of the concentration for each chemical, as a function of time.
                 Predicted ratio of compounds (enantiomeric ratio).
        """
        pass

    @staticmethod
    @abstractmethod
    def calculate_curves(data: pd.DataFrame) -> dict[str, pd.Series]:
        """
        Calculate the curves corresponding to the data (either experimental or predicted).
        The experimental curves will only be calculated once and are stored for subsequent use.
        Internally, the experimental and predicted curves will be compared against each other to determine the error.
        :param data: The data from which the curves should be calculated
        :return: dict containing a description of each curve, and the corresponding curve.
        """

    def calculate_error_functions(self, prediction: pd.DataFrame) -> pd.Series:
        """
        Calculate the error caused by each error function.
        :param prediction: The predicted concentrations.
        :param metric: The error metric which should be used, defaults to the mean absolute error.
        :return: The unweighted errors of each error function.
        """
        curves_predicted = self.calculate_curves(prediction)
        error = {}
        for curve_description, curve_prediction in curves_predicted.items():
            error[curve_description] = self.metric(
                y_true=self.experimental_curves[curve_description],
                y_pred=curve_prediction)

        return pd.Series(error)

    def weigh_errors(self, errors: pd.Series, ) -> pd.Series:
        """
        weighs the errors
        :param errors: unweighted errors
        :return: weighed errors
        """
        assert isinstance(errors, pd.Series)
        if self.weights is None:
            weights = np.ones(errors.shape)
            for description, weight in self.raw_weights.items():
                index = errors.index.str.contains(description)
                if len(index) == 0:
                    raise ValueError(f"no matches were found for {description}")
                weights[index] = weights[index] * weight
            self.weights = weights

        return errors * self.weights

    def calculate_total_error(self, errors: pd.Series) -> float:
        """
        weighs and sums the errors.
        :param errors: unweighted errors
        :return: weighed total error
        """
        return self.weigh_errors(errors).sum(skipna=False)

    def optimize(self,
                 x0: np.ndarray,
                 x_description: list[str],
                 bounds: list[tuple[float, float]],
                 path: str,
                 metadata: Optional[dict],
                 maxiter: float = 50000,
                 resume_from_simplex=None
                 ) -> None:
        """
        Optimizes the system, utilizing a nelder-mead algorithm.
        :param x0: Parameters which are to be optimized. Always contain the rate constants.
        :param x_description: Description of each parameter.
        :param bounds: A list containing tuples, which in turn contain the lower and upper bound for each parameter.
        :param path: Where the solution should be stored.
        :param metadata: The metadata that should be saved alongside the solution.
        :param maxiter: The maximum number of iterations.
        :param resume_from_simplex: When a simplex is given, the solution starts here.
        It can be used to resume the optimization process.
        """
        # enable logging of all information retrieved from the system
        log_path = f"{path}/optimization_log.json"
        if resume_from_simplex is None:  # new optimization progres
            logger = JSON_log(log_path)
            metadata_extended = {
                "raw_weights": self.raw_weights,
                "x0": x0,
                "x_description": x_description,
                "bounds": bounds,
                "maxiter": maxiter
            }
            if metadata is None:
                pass
            else:
                # overwrites the default meta data values
                for key, value in metadata.items():
                    metadata_extended[key] = value
            meta_data_log = JSON_log(f"{path}/settings_info.json", mode="new")
            meta_data_log.log(pd.Series(metadata_extended))
        else:
            logger = JSON_log(log_path, mode="append")

        def optimization_step(x):
            """The function is given a set of parameters by the Nelder-Mead algorithm.
            Proceeds to calculate the corresponding prediction and its total error.
            The results are stored in a log before the error is returned to the optimizer."""

            prediction, predicted_compound_ratio = self.create_prediction(x, x_description)
            errors = self.calculate_error_functions(prediction)
            total_error = self.calculate_total_error(errors)

            logger.log(pd.Series([x, total_error, predicted_compound_ratio], index=["x", "error", "ratio"]))
            return total_error

        def update_tqdm(_):
            """update the progress bar"""
            pbar.update(1)

        with tqdm(total=maxiter, miniters=25) as pbar:
            # the minimization process is stored within the log, containing all x's and errors.
            minimize(fun=optimization_step,
                     x0=x0,
                     method="Nelder-Mead",
                     bounds=bounds,
                     callback=update_tqdm,
                     options={"maxiter": maxiter, "disp": True, "adaptive": True, "return_all": False,
                              "initial_simplex": resume_from_simplex})

    @staticmethod
    def load_optimization_progress(path: str) -> OptimizerProgress:
        """
        Loads in the data from the log files.
        :param path: Folder in which the optimization_log.json and settings_info.json files can be found.
        :return: OptimizerProgress instance which contains all information that was logged.
        """
        return OptimizerProgress(path)
